require File.join(File.dirname(__FILE__), "service")

module Charrington
  class CreatePostgresTable
    # This service will create a table dynamically based on the JSON structure.
    # This is potentially called from Insert when an insert fails.

    @@timestamp_columns = %w(published_at sent_at original_timestamp received_at timestamp)

    include Service
    include LogStash::Util::Loggable
    attr_reader :connection, :event, :table_name, :columns, :schema, :opts, :transformer
    attr_accessor :column_types

    Error = Class.new(StandardError)
    CreateFailed = Class.new(Error)

    def initialize(connection, event, schema, table_name, columns, opts = {})
      @connection = connection
      @event = event.to_hash
      @table_name = table_name
      @schema = schema
      @columns = columns
      @transformer = opts[:transformer]
      @column_types = initial_columns
    end

    def call
      set_column_types
      create_table
      true
    rescue => e
      raise CreateFailed, e.message
    ensure
      @column_types.clear if @column_types.is_a? Array
    end

    private

    def set_column_types
      columns.each do |column|
        if @@timestamp_columns.include?(column)
          column_types << "#{column} TIMESTAMP"
          next
        end

        case event[column]
        when Time, LogStash::Timestamp
          column_types << "#{column} TIMESTAMP"
        when Date
          column_types << "#{column} DATE"
        when Integer
          column_types << "#{column} BIGINT"
        when BigDecimal
          column_types << "#{column} DECIMAL"
        when Float
          column_types << "#{column} DOUBLE PRECISION"
        when true, false
          column_types << "#{column} BOOLEAN"
        else
          column_types << "#{column} VARCHAR(512)"
        end
      end
    end

    def initial_columns
      if transformer == "postgres"
        [
          "id SERIAL PRIMARY KEY",
          "inserted_at TIMESTAMP DEFAULT NOW()"
        ]
      else
        [
          "uuid_ts TIMESTAMP DEFAULT NOW()"
        ]
      end
    end

    def create_table
      execute("CREATE TABLE IF NOT EXISTS #{schema}#{table_name} (#{column_types.join(', ')})")
    end

    def execute(sql)
      statement = connection.prepareStatement( sql.gsub(/\s+/, " ").strip )
      statement.execute()
    rescue Java::OrgPostgresqlUtil::PSQLException => e
      self.logger.error "PSQLException: #{e.message} sql=#{sql}"
    ensure
      statement.close unless statement.nil?
    end
  end
end
